/* -*- Mode:C; tab-width: 8 -*-
 * remote.c --- remote control of Netscape Navigator for Unix.
 * version 1.1.0, for Netscape Navigator 1.1 and newer.
 *
 * Copyright  1995 Netscape Communications Corporation, all rights reserved.
 * Created: Jamie Zawinski <jwz@netscape.com>, 24-Dec-94.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */


#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xmu/WinUtil.h>  /* for XmuClientWindow() */

/*#include "global_externs.h"*/

/* vroot.h is a header file which lets a client get along with `virtual root'
   window managers like swm, tvtwm, olvwm, etc.  If you don't have this header
   file, you can find it at "http://home.netscape.com/newsref/std/vroot.h".
   If you don't care about supporting virtual root window managers, you can
   comment this line out.
 */
#include "vroot.h"

#define MOZILLA_VERSION_PROP   "_MOZILLA_VERSION"
#define MOZILLA_LOCK_PROP      "_MOZILLA_LOCK"
#define MOZILLA_COMMAND_PROP   "_MOZILLA_COMMAND"
#define MOZILLA_RESPONSE_PROP  "_MOZILLA_RESPONSE"

static Atom XA_MOZILLA_VERSION  = 0;
static Atom XA_MOZILLA_LOCK     = 0;
static Atom XA_MOZILLA_COMMAND  = 0;
static Atom XA_MOZILLA_RESPONSE = 0;
short use_new_browser = 1;

static void
mozilla_remote_init_atoms (Display *dpy)
{
  if (! XA_MOZILLA_VERSION)
    XA_MOZILLA_VERSION = XInternAtom (dpy, MOZILLA_VERSION_PROP, False);
  if (! XA_MOZILLA_LOCK)
    XA_MOZILLA_LOCK = XInternAtom (dpy, MOZILLA_LOCK_PROP, False);
  if (! XA_MOZILLA_COMMAND)
    XA_MOZILLA_COMMAND = XInternAtom (dpy, MOZILLA_COMMAND_PROP, False);
  if (! XA_MOZILLA_RESPONSE)
    XA_MOZILLA_RESPONSE = XInternAtom (dpy, MOZILLA_RESPONSE_PROP, False);
}

static Window
mozilla_remote_find_window (Display *dpy)
{
  int i;
  Window root = RootWindowOfScreen (DefaultScreenOfDisplay (dpy));
  Window root2, parent, *kids;
  unsigned int nkids;
  Window result = 0;

  if (! XQueryTree (dpy, root, &root2, &parent, &kids, &nkids))
    return 0;

  if (root != root2)
    return 0;

  if (parent)
    return 0;

  if (! (kids && nkids))
    return 0;

  for (i = 0; i < nkids; i++)
    {
      Atom type;
      int format;
      unsigned long nitems, bytesafter;
      unsigned char *version = 0;
      Window w = XmuClientWindow (dpy, kids[i]);

      int status = XGetWindowProperty (dpy, w, XA_MOZILLA_VERSION,
                                       0, (65536 / sizeof (long)),
                                       False, XA_STRING,
                                       &type, &format, &nitems, &bytesafter,
                                       &version);
      if (! version)
        continue;

      XFree (version);

      if (status == Success && type != None)
        {
          result = w;
          break;
        }
    }
  
  return result;
}

static char *lock_data = 0;

static gboolean
mozilla_remote_obtain_lock (Display *dpy, 
                            Window window)
{
  Bool locked = False;
  Bool waited = False;

  if (!lock_data)
    {
      lock_data = (char *) malloc (255);
      sprintf (lock_data, "pid%d@", (int)getpid ());

      /*#if HAVE_GETHOSTNAME == 1*/
      if (gethostname (lock_data + strlen (lock_data), 100))
        {
          perror ("gethostname");
          return FALSE;
        }
      /*#else
#error You need the gethostname () function for this program
#endif */
    }

  do
    {
      int result;
      Atom actual_type;
      int actual_format;
      unsigned long nitems, bytes_after;
      unsigned char *data = 0;

      XGrabServer (dpy);   /* ################################# DANGER! */

      result = XGetWindowProperty (dpy, window, XA_MOZILLA_LOCK,
           0, (65536 / sizeof (long)),
           False, /* don't delete */
           XA_STRING,
           &actual_type, &actual_format,
           &nitems, &bytes_after,
           &data);
     
      if (result != Success || actual_type == None)
        {
          /* It's not now locked - lock it. */
          XChangeProperty (dpy, window, XA_MOZILLA_LOCK, XA_STRING, 8,
                           PropModeReplace, lock_data, strlen (lock_data));
          locked = True;
        }
      
      XUngrabServer (dpy); /* ################################# danger over */
      XSync (dpy, False);

      if (!locked)
        {
          /* We tried to grab the lock this time, and failed because someone
             else is holding it already.  So, wait for a PropertyDelete event
             to come in, and try again. */
          waited = True;

          g_warning ("window is locked, waiting...");

          while (1)
            {
              XEvent event;
              XNextEvent (dpy, &event);
              if (event.xany.type == DestroyNotify &&
                  event.xdestroywindow.window == window)
                {
                  g_warning ("Window was unexpectedly destroyed!");
                  return FALSE;
                }
              else if (event.xany.type == PropertyNotify 
                       && event.xproperty.state == PropertyDelete 
                       && event.xproperty.window == window 
                       && event.xproperty.atom == XA_MOZILLA_LOCK)
                {
                  /* Ok!  Someone deleted their lock, so now we can try
                     again. */
                  break;
                }
            }
        }

      if (data)
        XFree (data);
    }
  while (! locked);

  if (waited)
    g_warning ("Got lock...");

  return TRUE;
}


static gboolean
mozilla_remote_free_lock (Display *dpy, 
                          Window window)
{
  int result;
  Atom actual_type;
  int actual_format;
  unsigned long nitems, bytes_after;
  unsigned char *data = 0;
  gboolean retval = TRUE;

  result = XGetWindowProperty (dpy, window, XA_MOZILLA_LOCK,
             0, (65536 / sizeof (long)),
             True, /* atomic delete after */
             XA_STRING,
             &actual_type, &actual_format,
             &nitems, &bytes_after,
             &data);

  if (result != Success)
    retval = FALSE;
  else if (!data || !*data)
    retval = FALSE;
  else if (strcmp (data, lock_data))
    retval = FALSE;

  if (data)
    XFree (data);

  return retval;
}


static gint
mozilla_remote_command (Display *dpy, 
                        Window window, 
                        const gchar *command)
{
  gint result = 0;
  Bool done = False;
  gchar *new_command = 0;

  XChangeProperty (dpy, window, XA_MOZILLA_COMMAND, XA_STRING, 8,
                   PropModeReplace, command, strlen (command));

  while (!done)
    {
      XEvent event;
      XNextEvent (dpy, &event);
      if (event.xany.type == DestroyNotify 
          && event.xdestroywindow.window == window)
        {
          g_warning ("window 0x%x unexpectedly destroyed.",
                     (unsigned int) window);
          result = 6;
          goto DONE;
        }
      else if (event.xany.type == PropertyNotify 
               && event.xproperty.state == PropertyNewValue 
               && event.xproperty.window == window 
               && event.xproperty.atom == XA_MOZILLA_RESPONSE)
        {
          Atom actual_type;
          int actual_format;
          unsigned long nitems, bytes_after;
          unsigned char *data = 0;
          
          result = XGetWindowProperty (dpy, window, XA_MOZILLA_RESPONSE,
                                       0, (65536 / sizeof (long)),
                                       True, /* atomic delete after */
                                       XA_STRING,
                                       &actual_type, &actual_format,
                                       &nitems, &bytes_after,
                                       &data);
          
          if (result != Success)
            {
              g_warning ("failed reading " MOZILLA_RESPONSE_PROP
                         " from window 0x%0x.", (unsigned int) window);
              result = 6;
              done = True;
            }
          else if (!data || !*data)
            {
              g_warning ("invalid data on " MOZILLA_RESPONSE_PROP
                         " property of window 0x%0x.", 
                         (unsigned int) window);
              result = 6;
              done = True;
            }
          else if (*data == '1')  /* positive preliminary reply */
            {
              g_warning ("%s", data + 4);
              /* keep going */
              done = False;
            }
          else if (!strncmp (data, "200", 3))  /* positive completion */
            {
              result = 0;
              done = True;
            }
          else if (*data == '2')    /* positive completion */
            {
              g_warning ("%s", data + 4);
              result = 0;
              done = True;
            }
          else if (*data == '3')  /* positive intermediate reply */
            {
              g_warning ("server wants more information?  (%s)", data);
              result = 3;
              done = True;
            }
          else if (*data == '4' ||  /* transient negative completion */
                   *data == '5')  /* permanent negative completion */
            {
              g_warning (data + 4);
              result = (*data - '0');
              done = True;
            }
          else
            {
              g_warning ("unrecognised " MOZILLA_RESPONSE_PROP
                         " from window 0x%x: %s",
                         (unsigned int) window, data);
              result = 6;
              done = True;
            }
          
          if (data)
            XFree (data);
        }
    }

 DONE:

  if (new_command)
    free (new_command);

  return result;
}

/* this is patched to always add "noraise". this doesn't really seem to work, though.. */
gint
show_mozilla_url (gchar* url)
{
  int status = 0;
  gchar* command;
  Display *dpy;
  Window window;

  dpy = XOpenDisplay (NULL);
  if (dpy == NULL)
    return -1;

  mozilla_remote_init_atoms (dpy);

  window = mozilla_remote_find_window (dpy);
  if (window == 0)
    {
      status = -1;
      goto cleanup;
    }

  XSelectInput (dpy, window, (PropertyChangeMask|StructureNotifyMask));

  if (mozilla_remote_obtain_lock (dpy, window) == TRUE)
    {
      if (use_new_browser)
	command = g_strdup_printf ("openURL(%s, new-window, noraise)", url);
      else
	command = g_strdup_printf ("openURL(%s, noraise)", url);
  
      status = mozilla_remote_command (dpy, window, command);

      g_free (command);

      mozilla_remote_free_lock (dpy, window);
    }

 cleanup:
  XCloseDisplay (dpy);

  return status;
}
