#ifndef _OTF_HEADER_H_
#define _OTF_HEADER_H_

#include <stdio.h>

// definitions used in OTF s/w
#define OTFREV 1.1
#define HEADERSIZE 11264
#define MAXFILES 128
#define MAXCHAR 128
#define MAXSCANSIZE 1024
#define MAXHORNS 32
#define MAXDIM 256
#define MAXWIND 5
#define MAXORDER 5
#define NAN -10.0
#define RMAX 3.0
#define PI 3.1415926535;

#define DDUMMYS   19
#define RDUMMYS    9
#define IDUMMYS  104
#define MC_DDUMS  17
#define MC_RDUMS  49
#define MC_IDUMS  92
 
// error codes for OTF header reading and writing
#define OTFHEADER_FILE_OK 0
#define OTFHEADER_FILE_ERROR -1
#define OTFHEADER_REWIND_ERROR -2

/** structure for the OTF data file header
 */
struct OTFHeaderStruct
{
   /* 128 bytes */
   char otflab[16],floatopt[8],name[16],bunit[16],vlabel[16],
     xlabel[16],ylabel[16],cdum[24]; 
   /* 32*8=256 bytes */
   double utd,f0,fsky,azs,els,XCEN,YCEN,xvel1,xfr1,dveldin,dfrdin,fr1,fr2,
     ddum[DDUMMYS]; 
   /* 32*4=128 bytes */
   float tsid,utc,targetrms,tdump,tref,telrate,epoch,bw,vlsr,vclsr,vcsun,
     NOS,tau,h2omm,pamb,tamb,rotang,
     XSIZE,YSIZE,XOFFMAP,YOFFMAP,XRAMP,YRAMP,rdum[RDUMMYS];        
   /* 256*2=512 bytes */
   short int iscan,nhorns,nchan,nsample,nblock,blksize,telcoord,kx,ky,klo,khm,
	 harm,krecv,khorn,caltype,calscan,cfact,etab,kmap,raoff,decoff,fsrate,stype,
	 iconf,tsys[MAXHORNS],gain[MAXHORNS], 
     DXHORNS[MAXHORNS],DYHORNS[MAXHORNS],idum[IDUMMYS]; 
};
/** type for OTF data file header struct
 */
typedef struct OTFHeaderStruct OTFHeader_t; 

/** structure for the OTF position array 
*/
struct OTFPositionStruct
{
   /* 2048 bytes */
   short int ANGLE[MAXSCANSIZE];                  
   /* 8192 bytes */ 
   float XOFF[MAXSCANSIZE],YOFF[MAXSCANSIZE];      
};
/** type for OTF data file position struct
 */
typedef struct OTFPositionStruct OTFPositions_t; 

/** number of i*4's in a modcomp header
 */
#define MODELEMENTS 256

/** manages header information from ModComp to OTF data files
 */
class OTFHeader
{
 public:
  /** constructor
      @param name identification of shared memory region for header
      @param which identifies whether header for LOA (0) or LOB (1)
      @param n tells how many headers in the array
   */
  OTFHeader(char *name, int which, int n);
  /** destructor
   */
  ~OTFHeader();

  /** initialize the header data
   */
  void init();

  /** does a simple dump to stderr
   */
  void dump();

  /** allows user to load a couple key parameters for tests
      @param y the ut date in fractional years
      @param i the scan number
  */
  void setScan(double y, int i);

  /** creates a file name for OTF files from header info
      @param which indicated whether LOA (0) or LOB (1)
   */
  char *getFileName(int which);

  /** write the header data to an OPEN file
      @param outfile FILE pointer to open file for output.
      @return successful dump will return 0; errors given by negative numbers.
   */
  int write(FILE *outfile);

  /** read the header data from an OPEN file
      @param outfile FILE pointer to open file for output.
      @return successful read will return 0; negative numbers are errors
   */
  int read(FILE *outfile);

  /** updates the x and y offset and position angle of array.
      @param i map point counter
      @param x x offset in microradians from MC header
      @param y y offset in microradians from MC header
      @param a position angle in 10ths of milliradian from MC header
   */
  void updateOffset(int i, long x, long y, short a);


  /** return number of MAIN dumps in header
   */
  int getNSamples();

  /** return source name from header
   */
  char *getSourceName();

  /** return X offset of sample
      @param i index of sample to return; starts at 0
  */
  double getXOffset(int i);

  /** return Y offset of sample
      @param i index of sample to return; starts at 0
  */
  double getYOffset(int i);

  /** return paralactic angle of sample
      @param i index of sample to return; starts at 0
  */
  double getAngle(int i);

  /** load header data from ModComp data block
      @param h ModComp Header Block
   */
  void load(ModCompHeader_t h);

 protected:
  /** OTF Header Structure
   */
  OTFHeader_t *H;

};
#endif











